from typing import Iterable, Callable
from datetime import datetime, timezone
import pandas as pd

from discoverx.table_info import TableInfo

from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.window import Window
import pyspark.sql.types as T
import pyspark.sql.functions as F


class DeltaHousekeeping:

    def __init__(self, spark: SparkSession) -> None:
        self._spark = spark
        self.empty_schema = T.StructType(
            [
                T.StructField("catalog", T.StringType()),
                T.StructField("database", T.StringType()),
                T.StructField("tableName", T.StringType()),
            ]
        )

    @staticmethod
    def _process_describe_history(describe_detail_df: DataFrame, describe_history_df: DataFrame) -> pd.DataFrame:
        """
        processes the DESCRIBE HISTORY result of potentially several tables in different schemas/catalogs
        Provides
        - table stats (size and number of files)
        - timestamp for last & second last OPTIMIZE
        - stats of OPTIMIZE (including ZORDER)
        - timestamp for last & second last VACUUM

        returns a pandas DataFrame, and converts Spark internal dfs to pandas as soon as they are manageable
        the reason being that DESCRIBE HISTORY / DESCRIBE DETAIL cannot be cached
        """
        if not "operation" in describe_history_df.columns:
            return describe_detail_df.toPandas()

        # window over operation
        operation_order = describe_history_df.filter(F.col("operation").isin(["OPTIMIZE", "VACUUM END"])).withColumn(
            "operation_order",
            F.row_number().over(
                Window.partitionBy(["catalog", "database", "tableName", "operation"]).orderBy(F.col("timestamp").desc())
            ),
        )

        if operation_order.isEmpty():
            return describe_detail_df.toPandas()

        operation_order = operation_order.toPandas()

        # max & 2nd timestamp of OPTIMIZE into output
        out = describe_detail_df.toPandas().merge(
            operation_order[(operation_order.operation == "OPTIMIZE") & (operation_order.operation_order == 1)]
            .loc[:, ["catalog", "database", "tableName", "timestamp"]]
            .rename(columns={"timestamp": "max_optimize_timestamp"}),
            how="outer",
            on=["catalog", "database", "tableName"],
        )
        out = out.merge(
            operation_order[(operation_order.operation == "OPTIMIZE") & (operation_order.operation_order == 2)]
            .loc[:, ["catalog", "database", "tableName", "timestamp"]]
            .rename(columns={"timestamp": "2nd_optimize_timestamp"}),
            how="outer",
            on=["catalog", "database", "tableName"],
        )
        # max timestamp of VACUUM into output
        out = out.merge(
            operation_order[(operation_order.operation == "VACUUM END") & (operation_order.operation_order == 1)]
            .loc[:, ["catalog", "database", "tableName", "timestamp"]]
            .rename(columns={"timestamp": "max_vacuum_timestamp"}),
            how="outer",
            on=["catalog", "database", "tableName"],
        )
        out = out.merge(
            operation_order[(operation_order.operation == "VACUUM END") & (operation_order.operation_order == 2)]
            .loc[:, ["catalog", "database", "tableName", "timestamp"]]
            .rename(columns={"timestamp": "2nd_vacuum_timestamp"}),
            how="outer",
            on=["catalog", "database", "tableName"],
        )
        # summary of table metrics
        table_metrics_1 = operation_order[
            (operation_order["operation"] == "OPTIMIZE") & (operation_order["operation_order"] == 1)
        ].loc[:, ["catalog", "database", "tableName", "min_file_size", "p50_file_size", "max_file_size", "z_order_by"]]

        # write to output
        out = out.merge(table_metrics_1, how="outer", on=["catalog", "database", "tableName"])

        return out

    @staticmethod
    def save_as_table(
        result: DataFrame,
        housekeeping_table_name: str,
    ):
        """
        Static method to store intermediate results of the scan operation into Delta
        Would make sense only if using map_chunked from the `DataExplorer` object
        (otherwise tables are writen one by one into Delta with overhead)
        """
        (result.write.format("delta").mode("append").option("mergeSchema", "true").saveAsTable(housekeeping_table_name))

    def get_describe_detail(self, table_info: TableInfo):
        dd = self._spark.sql(
            f"""
            DESCRIBE DETAIL {table_info.catalog}.{table_info.schema}.{table_info.table};
        """
        )
        dd = (
            dd.withColumn("split", F.split(F.col("name"), "\."))
            .withColumn("catalog", F.col("split").getItem(0))
            .withColumn("database", F.col("split").getItem(1))
            .withColumn("tableName", F.col("split").getItem(2))
            .select(
                [
                    F.col("catalog"),
                    F.col("database"),
                    F.col("tableName"),
                    F.col("numFiles").alias("number_of_files"),
                    F.col("sizeInBytes").alias("bytes"),
                ]
            )
        )
        return dd

    @staticmethod
    def get_describe_history_statement(table_info: TableInfo):
        return f"""
                    SELECT 
                    '{table_info.catalog}' AS catalog,
                    '{table_info.schema}' AS database, 
                    '{table_info.table}' AS tableName, 
                    operation,
                    timestamp,
                    operationMetrics.minFileSize AS min_file_size,
                    operationMetrics.p50FileSize AS p50_file_size,
                    operationMetrics.maxFileSize AS max_file_size, 
                    operationParameters.zOrderBy AS z_order_by 
                    FROM (DESCRIBE HISTORY {table_info.catalog}.{table_info.schema}.{table_info.table})
                    WHERE operation in ('OPTIMIZE', 'VACUUM END')
                """

    def scan(
        self,
        table_info: TableInfo,
    ) -> pd.DataFrame:
        """
        Scans a table_info to fetch Delta stats
        - DESCRIBE DETAIL
        - DESCRIBE HISTORY
        """
        try:
            # runs a describe detail per table, figures out if exception
            dd = self.get_describe_detail(table_info)

            # prepares a DESCRIBE HISTORY statement per table (will be run outside the try-catch)
            statement = self.get_describe_history_statement(table_info)

            return self._process_describe_history(
                dd,
                self._spark.sql(statement),
            )

        except Exception as e:
            errors_df = self._spark.createDataFrame(
                [(table_info.catalog or "", table_info.schema, table_info.table, str(e))],
                ["catalog", "database", "tableName", "error"],
            )
            return errors_df.toPandas()


class DeltaHousekeepingActions:
    """
    Processes the output of the `DeltaHousekeeping` object to provide recommendations
    - tables that need to be OPTIMIZED/VACUUM'ed
    - are tables OPTIMIZED/VACUUM'ed often enough
    - tables that have small files / tables for which ZORDER is not being effective
    """

    def __init__(
        self,
        mapped_pd_dfs: Iterable[pd.DataFrame],
        spark: SparkSession = None,
        min_table_size_optimize: int = 128 * 1024 * 1024,  # i.e. 128 MB
        min_days_not_optimized: int = 7,  # in days
        min_days_not_vacuumed: int = 31,  # in days
        max_optimize_freq: int = 2,  # in days - e.g. 2 means that a daily run would be flagged
        max_vacuum_freq: int = 2,  # in days - e.g. 2 means that a daily run would be flagged
        small_file_threshold: int = 32 * 1024 * 1024,  # i.e. 32 MB
        min_number_of_files_for_zorder: int = 8,
        stats: pd.DataFrame = None,  # for testability only
    ) -> None:
        if spark is None:
            spark = SparkSession.builder.getOrCreate()
        self._spark = spark

        if stats is None:
            self._mapped_pd_dfs = mapped_pd_dfs
            stats = pd.concat(self._mapped_pd_dfs)
        self._stats_df: DataFrame = self._spark.createDataFrame(stats)
        for column in [col_name for col_name in self._stats_df.columns if "timestamp" in col_name]:
            column_type = [dtype for col_name, dtype in self._stats_df.dtypes if col_name == column][0]
            if column_type == "timestamp":
                continue
            self._stats_df = self._stats_df.withColumn(
                column, F.when(F.isnan(F.col(column)), None).otherwise(F.col(column))
            )
        self._stats_rec: DataFrame = None

        self.min_table_size_optimize = min_table_size_optimize
        self.min_days_not_optimized = min_days_not_optimized
        self.min_days_not_vacuumed = min_days_not_vacuumed
        self.max_optimize_freq = max_optimize_freq
        self.max_vacuum_freq = max_vacuum_freq
        self.small_file_threshold = small_file_threshold
        self.min_number_of_files_for_zorder = min_number_of_files_for_zorder
        self.reason_col_suffix = "_reason"
        self.recomendations_dict = {
            "not_optimized": {
                "legend": "Tables that have never been OPTIMIZED and would benefit from it",
                "col_name": "rec_not_optimized",
            },
            "not_vacuumed": {"legend": "Tables that have never been VACUUM'ed", "col_name": "rec_not_vacuumed"},
            "not_optimized_last_days": {
                "legend": "Tables that are not OPTIMIZED often enough",
                "col_name": "rec_not_optimized_last_days",
            },
            "not_vacuumed_last_days": {
                "legend": "Tables that are not VACUUM'ed often enough",
                "col_name": "rec_not_vacuumed_last_days",
            },
            "optimized_too_freq": {
                "legend": "Tables that are OPTIMIZED too often",
                "col_name": "rec_optimized_too_freq",
            },
            "vacuumed_too_freq": {"legend": "Tables that are VACUUM'ed too often", "col_name": "rec_vacuumed_too_freq"},
            "do_not_need_optimize": {
                "legend": "Tables that are too small to be OPTIMIZED",
                "col_name": "rec_do_not_need_optimize",
            },
            "to_analyze": {"legend": "Tables that need more analysis -small_files", "col_name": "rec_to_analyze"},
            "zorder_not_effective": {
                "legend": "Tables for which ZORDER is not being effective",
                "col_name": "rec_zorder_not_effective",
            },
        }

    def _need_optimize(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["not_optimized"]
        return stats_rec.withColumn(
            conf_dict["col_name"],
            F.when(
                F.col("max_optimize_timestamp").isNull()
                & F.col("bytes").isNotNull()
                & (F.col("bytes").astype("int") > F.lit(self.min_table_size_optimize)),
                F.lit(True),
            ).otherwise(F.lit(False)),
        )

    def _optimize_not_needed(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["do_not_need_optimize"]
        return stats_rec.withColumn(
            conf_dict["col_name"],
            F.when(
                F.col("max_optimize_timestamp").isNotNull()
                & F.col("bytes").isNotNull()
                & (F.col("bytes").astype("int") < F.lit(self.min_table_size_optimize)),
                F.lit(True),
            ).otherwise(F.lit(False)),
        )

    def _not_optimized_last_days(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["not_optimized_last_days"]
        return stats_rec.withColumn(
            "optimize_lag", F.date_diff(F.lit(datetime.today()), F.col("max_optimize_timestamp"))
        ).withColumn(
            conf_dict["col_name"],
            F.when(F.col("optimize_lag") > F.lit(self.min_days_not_optimized), F.lit(True)).otherwise(F.lit(False)),
        )

    def _optimized_too_frequently(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["optimized_too_freq"]
        return stats_rec.withColumn(
            "optimize_freq",
            F.when(
                F.col("max_optimize_timestamp").isNotNull() & F.col("2nd_optimize_timestamp").isNotNull(),
                F.date_diff(F.col("max_optimize_timestamp"), F.col("2nd_optimize_timestamp")),
            ),
        ).withColumn(
            conf_dict["col_name"],
            F.when(F.col("optimize_freq") < F.lit(self.max_optimize_freq), F.lit(True)).otherwise(F.lit(False)),
        )

    def _never_vacuumed(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["not_vacuumed"]
        return stats_rec.withColumn(
            conf_dict["col_name"], F.when(F.col("max_vacuum_timestamp").isNull(), F.lit(True)).otherwise(F.lit(False))
        )

    def _not_vacuumed_last_days(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["not_vacuumed_last_days"]
        return stats_rec.withColumn(
            "vacuum_lag", F.date_diff(F.lit(datetime.today()), F.col("max_vacuum_timestamp"))
        ).withColumn(
            conf_dict["col_name"],
            F.when(F.col("vacuum_lag") > F.lit(self.min_days_not_vacuumed), F.lit(True)).otherwise(F.lit(False)),
        )

    def _vacuumed_too_frequently(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["vacuumed_too_freq"]
        return stats_rec.withColumn(
            "vacuum_freq",
            F.when(
                F.col("max_vacuum_timestamp").isNotNull() & F.col("2nd_vacuum_timestamp").isNotNull(),
                F.date_diff(F.col("max_vacuum_timestamp"), F.col("2nd_vacuum_timestamp")),
            ),
        ).withColumn(
            conf_dict["col_name"],
            F.when(F.col("vacuum_freq") < F.lit(self.max_vacuum_freq), F.lit(True)).otherwise(F.lit(False)),
        )

    def _analyze_these_tables(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["to_analyze"]
        return stats_rec.withColumn(
            conf_dict["col_name"],
            F.when(
                F.col("max_optimize_timestamp").isNotNull()
                & F.col("p50_file_size").isNotNull()
                & (F.col("number_of_files") > F.lit(1))
                & (F.col("p50_file_size").astype("int") < F.lit(self.small_file_threshold)),
                F.lit(True),
            ).otherwise(F.lit(False)),
        )

    def _zorder_not_effective(self, stats_rec: DataFrame) -> DataFrame:
        conf_dict = self.recomendations_dict["zorder_not_effective"]
        return (
            stats_rec.withColumn(
                "z_order_by_clean",
                F.when(
                    F.col("max_optimize_timestamp").isNull()
                    | F.col("p50_file_size").isNull()
                    | (F.col("z_order_by") == "[]"),
                    None,
                ).otherwise(
                    F.regexp_replace(
                        F.regexp_replace(F.regexp_replace(F.col("z_order_by"), "\\[", ""), "\\]", ""), '"', ""
                    )
                ),
            )
            .withColumn("z_order_by_array", F.split(F.col("z_order_by_clean"), ","))
            .withColumn(
                conf_dict["col_name"],
                F.when(
                    (F.size(F.col("z_order_by_array")) > 0)
                    & (F.col("number_of_files") < F.lit(self.min_number_of_files_for_zorder)),
                    F.lit(True),
                ).otherwise(F.lit(False)),
            )
        )

    def display(self) -> None:
        """Executes the Delta housekeeping analysis and displays a sample of results"""
        return self.apply().display()

    def apply(self) -> DataFrame:
        """Displays recommendations in a DataFrame format"""
        return self.generate_recommendations()

    def generate_recommendations(self) -> DataFrame:
        """
        Generates Delta Housekeeping recommendations as a list of dictionaries (internal use + unit tests only)
        A dict per recommendation where:
        - The key is the legend of the recommendation
        - The value is a pandas df with the affected tables
        """
        if self._stats_rec is None:
            stats_rec = self._need_optimize(self._stats_df)
            stats_rec = self._never_vacuumed(stats_rec)
            stats_rec = self._not_optimized_last_days(stats_rec)
            stats_rec = self._not_vacuumed_last_days(stats_rec)
            stats_rec = self._optimized_too_frequently(stats_rec)
            stats_rec = self._vacuumed_too_frequently(stats_rec)
            stats_rec = self._optimize_not_needed(stats_rec)
            stats_rec = self._analyze_these_tables(stats_rec)
            stats_rec = self._zorder_not_effective(stats_rec)
            self._stats_rec = stats_rec

        return self._stats_rec

    def _explain(self) -> Iterable[dict]:
        stats = self.generate_recommendations()
        out = []
        for _, item in self.recomendations_dict.items():
            legend = None
            col_name = None
            for k, v in item.items():
                if k == "legend":
                    legend = v
                elif k == "col_name":
                    col_name = v

            out.append({legend: stats.filter(F.col(col_name))})

        return out

    def explain(self) -> None:
        from databricks.sdk.runtime import display

        for item in self._explain():
            for legend, df in item.items():
                display(legend)
                display(df)



from typing import Optional
from discoverx.common import helper
from pyspark.sql.types import Row
from dataclasses import dataclass


@dataclass
class ColumnInfo:
    name: str
    data_type: str
    partition_index: int
    classes: list[str]


@dataclass
class TagInfo:
    tag_name: str
    tag_value: str


@dataclass
class ColumnTagInfo:
    column_name: str
    tag_name: str
    tag_value: str


@dataclass
class TagsInfo:
    column_tags: Optional[list[ColumnTagInfo]]
    table_tags: Optional[list[TagInfo]]
    schema_tags: Optional[list[TagInfo]]
    catalog_tags: Optional[list[TagInfo]]


@dataclass
class TableInfo:
    catalog: Optional[str]
    schema: str
    table: str
    columns: list[ColumnInfo]
    tags: Optional[TagsInfo]

    def get_columns_by_class(self, class_name: str):
        return [ClassifiedColumn(col.name, class_name) for col in self.columns if class_name in col.classes]


@dataclass
class ClassifiedColumn:
    name: str
    class_name: str


class InfoFetcher:
    def __init__(self, spark, information_schema="system.information_schema") -> None:
        self.information_schema = information_schema
        self.spark = spark

    @staticmethod
    def _get_tag_info(row: Row) -> TagsInfo:
        if all(item in row.asDict().keys() for item in ["column_tags", "table_tags", "schema_tags", "catalog_tags"]):
            if row["column_tags"] is None:
                column_tags = []
            else:
                column_tags = [
                    ColumnTagInfo(tag["column_name"], tag["tag_name"], tag["tag_value"]) for tag in row["column_tags"]
                ]

            if row["table_tags"] is None:
                table_tags = []
            else:
                table_tags = [TagInfo(tag["tag_name"], tag["tag_value"]) for tag in row["table_tags"]]

            if row["schema_tags"] is None:
                schema_tags = []
            else:
                schema_tags = [TagInfo(tag["tag_name"], tag["tag_value"]) for tag in row["schema_tags"]]

            if row["catalog_tags"] is None:
                catalog_tags = []
            else:
                catalog_tags = [TagInfo(tag["tag_name"], tag["tag_value"]) for tag in row["catalog_tags"]]

            tags = TagsInfo(column_tags, table_tags, schema_tags, catalog_tags)
        else:
            tags = None
        return tags

    @staticmethod
    def _to_info_row(row: Row) -> TableInfo:
        columns = [
            ColumnInfo(col["column_name"], col["data_type"], col["partition_index"], []) for col in row["table_columns"]
        ]

        return TableInfo(
            row["table_catalog"],
            row["table_schema"],
            row["table_name"],
            columns=columns,
            tags=InfoFetcher._get_tag_info(row),
        )

    def _to_info_list(self, info_rows: list[Row]) -> list[TableInfo]:
        filtered_tables = [self._to_info_row(row) for row in info_rows]
        return filtered_tables

    def get_tables_info(
        self,
        catalogs: str,
        schemas: str,
        tables: str,
        columns: list[str] = [],
        with_tags=False,
        data_source_formats: list[str] = ["DELTA"],
    ) -> list[TableInfo]:
        # Filter tables by matching filter
        table_list_sql = self._get_table_list_sql(catalogs, schemas, tables, columns, with_tags, data_source_formats)

        filtered_tables = self.spark.sql(table_list_sql).collect()

        if len(filtered_tables) == 0:
            raise ValueError(f"No tables found matching filter: {catalogs}.{schemas}.{tables}")

        return self._to_info_list(filtered_tables)

    def _get_table_list_sql(
        self,
        catalogs: str,
        schemas: str,
        tables: str,
        columns: list[str] = [],
        with_tags=False,
        data_source_formats: list[str] = ["DELTA"],
    ) -> str:
        """
        Returns a SQL expression which returns a list of columns matching
        the specified filters

        Returns:
            string: The SQL expression
        """

        if "*" in catalogs:
            catalog_sql = f"""AND regexp_like(table_catalog, "^{catalogs.replace("*", ".*")}$")"""
            catalog_tags_sql = f"""AND regexp_like(catalog_name, "^{catalogs.replace("*", ".*")}$")"""
        else:
            catalog_sql = f"""AND table_catalog = "{catalogs}" """
            catalog_tags_sql = f"""AND catalog_name = "{catalogs}" """

        if "*" in schemas:
            schema_sql = f"""AND regexp_like(table_schema, "^{schemas.replace("*", ".*")}$")"""
            schema_tags_sql = f"""AND regexp_like(schema_name, "^{schemas.replace("*", ".*")}$")"""
        else:
            schema_sql = f"""AND table_schema = "{schemas}" """
            schema_tags_sql = f"""AND schema_name = "{schemas}" """

        if "*" in tables:
            table_sql = f"""AND regexp_like(table_name, "^{tables.replace("*", ".*")}$")"""
        else:
            table_sql = f"""AND table_name = "{tables}" """

        if columns:
            match_any_col = "|".join([f'({c.replace("*", ".*")})' for c in columns])
            columns_sql = f"""AND regexp_like(column_name, "^{match_any_col}$")"""

        data_source_formats_values = ",".join("'{0}'".format(f).upper() for f in data_source_formats)

        with_column_info_sql = f"""
        WITH all_user_tbl_list AS (
            SELECT DISTINCT
                table_catalog, 
                table_schema, 
                table_name
            FROM {self.information_schema}.columns
            WHERE 
                table_schema != "information_schema" 
                {catalog_sql if catalogs != "*" else ""}
                {schema_sql if schemas != "*" else ""}
                {table_sql if tables != "*" else ""}
                {columns_sql if columns else ""}
        ),
        
        req_tbl_list AS (
        SELECT DISTINCT
            table_catalog,
            table_schema,
            table_name
            FROM {self.information_schema}.tables
            WHERE
                table_schema != "information_schema"
                {catalog_sql if catalogs != "*" else ""}
                {schema_sql if schemas != "*" else ""}
                {table_sql if tables != "*" else ""}
                and table_type in ('MANAGED','EXTERNAL','MANAGED_SHALLOW_CLONE','EXTERNAL_SHALLOW_CLONE')
                and data_source_format in ({data_source_formats_values})
        ),
        
        filtered_tbl_list AS (
            SELECT a.* 
            FROM all_user_tbl_list a
            INNER JOIN
            req_tbl_list r ON(
            a.table_catalog <=> r.table_catalog AND
            a.table_schema = r.table_schema AND
            a.table_name = r.table_name
            )
        
        ),

        col_list AS (
          SELECT
            info_schema.table_catalog,
            info_schema.table_schema,
            info_schema.table_name,
            collect_list(struct(column_name, data_type, partition_index)) as table_columns
          FROM {self.information_schema}.columns info_schema
          WHERE 
                table_schema != "information_schema" 
                {catalog_sql if catalogs != "*" else ""}
                {schema_sql if schemas != "*" else ""}
                {table_sql if tables != "*" else ""}
          GROUP BY info_schema.table_catalog, info_schema.table_schema, info_schema.table_name
        ),

        with_column_info AS (
          SELECT
              col_list.*
          FROM col_list
          INNER JOIN filtered_tbl_list ON (
              col_list.table_catalog <=> filtered_tbl_list.table_catalog AND
              col_list.table_schema = filtered_tbl_list.table_schema AND
              col_list.table_name = filtered_tbl_list.table_name)
        )

        """

        tags_sql = f"""
        ,
        catalog_tags AS (
          SELECT
            info_schema.catalog_name AS table_catalog,
            collect_list(struct(tag_name, tag_value)) as catalog_tags
          FROM {self.information_schema}.catalog_tags info_schema
          WHERE 
                catalog_name != "system"
                {catalog_tags_sql if catalogs != "*" else ""}
          GROUP BY info_schema.catalog_name
        ),

        schema_tags AS (
          SELECT
            info_schema.catalog_name AS table_catalog,
            info_schema.schema_name AS table_schema,
            collect_list(struct(tag_name, tag_value)) as schema_tags
          FROM {self.information_schema}.schema_tags info_schema
          WHERE 
                schema_name != "information_schema" 
                {catalog_tags_sql if catalogs != "*" else ""}
                {schema_tags_sql if schemas != "*" else ""}
          GROUP BY info_schema.catalog_name, info_schema.schema_name
        ),

        table_tags AS (
          SELECT
            info_schema.catalog_name AS table_catalog,
            info_schema.schema_name AS table_schema,
            info_schema.table_name,
            collect_list(struct(tag_name, tag_value)) as table_tags
          FROM {self.information_schema}.table_tags info_schema
          WHERE 
                schema_name != "information_schema" 
                {catalog_tags_sql if catalogs != "*" else ""}
                {schema_tags_sql if schemas != "*" else ""}
                {table_sql if tables != "*" else ""}
          GROUP BY info_schema.catalog_name, info_schema.schema_name, info_schema.table_name
        ),

        column_tags AS (
          SELECT
            info_schema.catalog_name AS table_catalog,
            info_schema.schema_name AS table_schema,
            info_schema.table_name,
            collect_list(struct(column_name, tag_name, tag_value)) as column_tags
          FROM {self.information_schema}.column_tags info_schema
          WHERE 
                schema_name != "information_schema" 
                {catalog_tags_sql if catalogs != "*" else ""}
                {schema_tags_sql if schemas != "*" else ""}
                {table_sql if tables != "*" else ""}
          GROUP BY info_schema.catalog_name, info_schema.schema_name, info_schema.table_name
        ),

        tags AS (
          SELECT 
            filtered_tbl_list.table_catalog,
            filtered_tbl_list.table_schema,
            filtered_tbl_list.table_name,
            catalog_tags.catalog_tags,
            schema_tags.schema_tags,
            table_tags.table_tags,
            column_tags.column_tags
          FROM filtered_tbl_list
          LEFT OUTER JOIN table_tags ON (
            table_tags.table_catalog <=> filtered_tbl_list.table_catalog AND 
            table_tags.table_schema = filtered_tbl_list.table_schema AND 
            table_tags.table_name = filtered_tbl_list.table_name
            )
          LEFT OUTER JOIN schema_tags
          ON filtered_tbl_list.table_catalog <=> schema_tags.table_catalog AND filtered_tbl_list.table_schema = schema_tags.table_schema 
          LEFT OUTER JOIN column_tags
          ON filtered_tbl_list.table_catalog <=> column_tags.table_catalog AND filtered_tbl_list.table_schema = column_tags.table_schema AND filtered_tbl_list.table_name = column_tags.table_name
          LEFT OUTER JOIN catalog_tags
          ON catalog_tags.table_catalog <=> filtered_tbl_list.table_catalog
        )

        
        """

        if with_tags:
            sql = (
                with_column_info_sql
                + tags_sql
                + f"""
            SELECT
                with_column_info.*,
                tags.table_tags,
                tags.catalog_tags,
                tags.schema_tags,
                tags.table_tags,
                tags.column_tags
            FROM with_column_info
            LEFT OUTER JOIN tags ON (
              with_column_info.table_catalog <=> tags.table_catalog AND
              with_column_info.table_schema = tags.table_schema AND
              with_column_info.table_name = tags.table_name)
            """
            )
        else:
            sql = (
                with_column_info_sql
                + f"""
                SELECT
                *
                FROM with_column_info
                """
            )

        return helper.strip_margin(sql)



import concurrent.futures
import copy
import re
import pandas as pd
from typing import Optional, List, Callable, Iterable, Any
from discoverx import logging
from discoverx.common import helper
from discoverx.discovery import Discovery
from discoverx.rules import Rule
from discoverx.table_info import TagsInfo, ColumnTagInfo, TagInfo, ColumnInfo, TableInfo
from functools import reduce
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import lit
from discoverx.table_info import InfoFetcher, TableInfo
from discoverx.delta_housekeeping import DeltaHousekeeping, DeltaHousekeepingActions


logger = logging.Logging()


class DataExplorer:
    FROM_COMPONENTS_EXPR = r"^(([0-9a-zA-Z_\*-]+)\.([0-9a-zA-Z_\*-]+)\.([0-9a-zA-Z_\*-]+))$"

    def __init__(self, from_tables, spark: SparkSession, info_fetcher: InfoFetcher) -> None:
        self._from_tables = from_tables
        (
            self._catalogs,
            self._schemas,
            self._tables,
        ) = DataExplorer.validate_from_components(from_tables)
        self._spark = spark
        self._info_fetcher = info_fetcher
        self._having_columns = []
        self._sql_query_template = None
        self._max_concurrency = 10
        self._with_tags = False
        self._data_source_formats = ["DELTA"]

    @staticmethod
    def validate_from_components(from_tables: str):
        """Extracts the catalog, schema and table name from the from_table string"""
        matches = re.findall(DataExplorer.FROM_COMPONENTS_EXPR, from_tables)
        if len(matches) == 1 and len(matches[0]) == 4:
            return (matches[0][1], matches[0][2], matches[0][3])
        else:
            raise ValueError(
                f"Invalid from_tables statement '{from_tables}'. Should be a string in format 'table_catalog.table_schema.table_name'. You can use '*' as wildcard."
            )

    def __deepcopy__(self, memo):
        new_obj = type(self).__new__(self.__class__)
        new_obj.__dict__.update(self.__dict__)
        new_obj._catalogs = copy.deepcopy(self._catalogs)
        new_obj._schemas = copy.deepcopy(self._schemas)
        new_obj._tables = copy.deepcopy(self._tables)
        new_obj._having_columns = copy.deepcopy(self._having_columns)
        new_obj._sql_query_template = copy.deepcopy(self._sql_query_template)
        new_obj._max_concurrency = copy.deepcopy(self._max_concurrency)
        new_obj._with_tags = copy.deepcopy(self._with_tags)
        new_obj._data_source_formats = copy.deepcopy(self._data_source_formats)

        new_obj._spark = self._spark
        new_obj._info_fetcher = self._info_fetcher

        return new_obj

    def having_columns(self, *columns) -> "DataExplorer":
        """Will select tables that contain any of the provided columns

        Args:
            columns (list[str]): The list of column names to filter by
        """
        new_obj = copy.deepcopy(self)
        new_obj._having_columns.extend(columns)
        return new_obj

    def with_data_source_formats(self, data_source_formats: list[str] = ["DELTA"]) -> "DataExplorer":
        """Filter tables with provided data source formats. Defaults to DELTA only. Possible Values 'DELTA', 'CSV', 'JSON', 'PARQUET', 'TEXT', 'ORC' etc

        Args:
            data_source_formats (list[str]): The list of data source formats to filter by
        """
        new_obj = copy.deepcopy(self)
        new_obj._data_source_formats = data_source_formats
        return new_obj

    def with_concurrency(self, max_concurrency) -> "DataExplorer":
        """Sets the maximum number of concurrent queries to run"""
        new_obj = copy.deepcopy(self)
        new_obj._max_concurrency = max_concurrency
        return new_obj

    def with_tags(self, use_tags=True) -> "DataExplorer":
        """Defines if tags should be collected when getting table metadata"""
        new_obj = copy.deepcopy(self)
        new_obj._with_tags = use_tags
        return new_obj

    def with_sql(self, sql_query_template: str) -> "DataExplorerActions":
        """Sets the SQL query template to use for the data exploration

        Args:
            sql_query_template (str): The SQL query template to use. The template might contain the following variables:
                - table_catalog: The table catalog name
                - table_schema: The table schema name
                - table_name: The table name
                - full_table_name: The full table name (catalog.schema.table)
                - stack_string_columns: A SQL expression that returns a table with two columns: column_name and string_value
        """
        return self.apply_sql(sql_query_template)

    def apply_sql(self, sql_query_template: str) -> "DataExplorerActions":
        """[DEPRECATED] Sets the SQL query template to use for the data exploration

        Args:
            sql_query_template (str): The SQL query template to use. The template might contain the following variables:
                - table_catalog: The table catalog name
                - table_schema: The table schema name
                - table_name: The table name
                - full_table_name: The full table name (catalog.schema.table)
                - stack_string_columns: A SQL expression that returns a table with two columns: column_name and string_value
        """
        new_obj = copy.deepcopy(self)
        new_obj._sql_query_template = sql_query_template
        return DataExplorerActions(new_obj, spark=self._spark, info_fetcher=self._info_fetcher)

    def unpivot_string_columns(self, sample_size=None) -> "DataExplorerActions":
        """Returns a DataExplorerActions object that will run a query that will melt all string columns into a pair of columns (column_name, string_value)

        Args:
            sample_size (int, optional): The number of rows to sample. Defaults to None (Return all rows).
        """

        sql_query_template = """
        SELECT
            {stack_string_columns} AS (column_name, string_value)
        FROM {full_table_name}
        """
        if sample_size is not None:
            sql_query_template += f"TABLESAMPLE ({sample_size} ROWS)"

        return self.with_sql(sql_query_template)

    def unpivot_all_columns(self, sample_size=None) -> "DataExplorerActions":
        """Returns a DataExplorerActions object that will run a query that will melt all columns into a pair of columns (column_name, string_value)

        Args:
            sample_size (int, optional): The number of rows to sample. Defaults to None (Return all rows).
        """

        sql_query_template = """
        SELECT
            {stack_all_columns_as_string} AS (column_name, string_value)
        FROM {full_table_name}
        """
        if sample_size is not None:
            sql_query_template += f"TABLESAMPLE ({sample_size} ROWS)"

        return self.with_sql(sql_query_template)

    def scan(
        self,
        rules="*",
        sample_size=10000,
        what_if: bool = False,
        custom_rules: Optional[List[Rule]] = None,
        locale: str = None,
    ):
        discover = Discovery(
            self._spark,
            self._catalogs,
            self._schemas,
            self._tables,
            self._info_fetcher.get_tables_info(
                catalogs=self._catalogs,
                schemas=self._schemas,
                tables=self._tables,
                columns=self._having_columns,
                data_source_formats=self._data_source_formats,
            ),
            custom_rules=custom_rules,
            locale=locale,
        )
        discover.scan(rules=rules, sample_size=sample_size, what_if=what_if)
        return discover

    def map(self, f: Callable[[TableInfo], Any]) -> list[Any]:
        """Runs a function for each table in the data explorer

        Args:
            f (function): The function to run. The function should accept a TableInfo object as input and return any object as output.

        Returns:
            list[any]: A list of the results of running the function for each table
        """
        res = []
        table_list = self._info_fetcher.get_tables_info(
            self._catalogs,
            self._schemas,
            self._tables,
            self._having_columns,
            self._with_tags,
            self._data_source_formats,
        )
        with concurrent.futures.ThreadPoolExecutor(max_workers=self._max_concurrency) as executor:
            # Submit tasks to the thread pool
            futures = [executor.submit(f, table_info) for table_info in table_list]

            # Process completed tasks
            for future in concurrent.futures.as_completed(futures):
                result = future.result()
                if result is not None:
                    res.append(result)

        logger.debug("Finished lakehouse map task")

        return res

    def delta_housekeeping(self) -> pd.DataFrame:
        """
        Gathers stats and recommendations on Delta Housekeeping
        """
        dh = DeltaHousekeeping(self._spark)
        dfs_pd: Iterable[pd.DataFrame] = self.map(dh.scan)
        return DeltaHousekeepingActions(dfs_pd, spark=self._spark)


class DataExplorerActions:
    def __init__(
        self,
        data_explorer: DataExplorer,
        spark: SparkSession = None,
        info_fetcher: InfoFetcher = None,
    ) -> None:
        self._data_explorer = data_explorer
        if spark is None:
            spark = SparkSession.builder.getOrCreate()
        if info_fetcher is None:
            info_fetcher = InfoFetcher(spark)
        self._info_fetcher = info_fetcher
        self._spark = spark

    @staticmethod
    def _get_stack_string_columns_expression(table_info: TableInfo) -> str:
        string_col_names = [c.name for c in table_info.columns if c.data_type.lower() == "string"]
        stack_parameters = ", ".join([f"'{c}', `{c}`" for c in string_col_names])
        return f"stack({len(string_col_names)}, {stack_parameters})"

    @staticmethod
    def _get_stack_all_columns_expression(table_info: TableInfo) -> str:
        col_names = [c.name for c in table_info.columns]
        stack_parameters = ", ".join([f"'{c}', cast(`{c}` AS string)" for c in col_names])
        return f"stack({len(col_names)}, {stack_parameters})"

    @staticmethod
    def _build_sql(sql_template: str, table_info: TableInfo) -> str:
        if table_info.catalog and table_info.catalog != "None":
            full_table_name = f"{table_info.catalog}.{table_info.schema}.{table_info.table}"
        else:
            full_table_name = f"{table_info.schema}.{table_info.table}"

        stack_string_columns = DataExplorerActions._get_stack_string_columns_expression(table_info)
        stack_all_columns_as_string = DataExplorerActions._get_stack_all_columns_expression(table_info)

        sql = sql_template.format(
            table_catalog=table_info.catalog,
            table_schema=table_info.schema,
            table_name=table_info.table,
            full_table_name=full_table_name,
            stack_string_columns=stack_string_columns,
            stack_all_columns_as_string=stack_all_columns_as_string,
        )
        return sql

    def _run_sql(self, sql: str, table_info: TableInfo) -> DataFrame:
        logger.debug(f"Running SQL query: {sql}")

        try:
            df = (
                self._spark.sql(sql)
                .withColumn("table_catalog", lit(table_info.catalog))
                .withColumn("table_schema", lit(table_info.schema))
                .withColumn("table_name", lit(table_info.table))
            )
            logger.debug(f"Finished running SQL query: {sql}")
            return df
        except Exception as e:
            logger.error(f"Error running SQL query for: {table_info.catalog}.{table_info.schema}.{table_info.table}.")
            logger.error(e)
            return None

    def _get_sql_commands(self, data_explorer: DataExplorer) -> list[tuple[str, TableInfo]]:
        logger.debug("Launching lakehouse scanning task\n")

        table_list = self._info_fetcher.get_tables_info(
            data_explorer._catalogs,
            data_explorer._schemas,
            data_explorer._tables,
            data_explorer._having_columns,
            data_explorer._with_tags,
        )
        sql_commands = [
            (
                DataExplorerActions._build_sql(data_explorer._sql_query_template, table),
                table,
            )
            for table in table_list
        ]
        return sql_commands

    def explain(self) -> None:
        """Prints a friendly explanation of the data exploration that will be performed

        The explanation will include the SQL query that will be executed, and the tables that will be scanned
        """
        column_filter_explanation = ""
        sql_explanation = ""

        if self._data_explorer._having_columns:
            column_filter_explanation = (
                f"only for tables that have all the following columns: {self._data_explorer._having_columns}"
            )
        if self._data_explorer._sql_query_template:
            sql_explanation = f"The SQL to be executed is (just a moment, generating it...):"

        explanation = f"""
        DiscoverX will apply the following SQL template

        {self._data_explorer._sql_query_template}

        to the tables in the following catalog, schema, table combinations:
        {self._data_explorer._from_tables}
        {column_filter_explanation}
        {sql_explanation}
        """
        logger.friendly(helper.strip_margin(explanation))

        detailed_explanation = ""
        if self._data_explorer._sql_query_template:
            sql_commands = self._get_sql_commands(self._data_explorer)
            for sql, table in sql_commands:
                detailed_explanation += f"""
                <p>For table: {table.catalog}.{table.schema}.{table.table}</p>
                <pre><code>{sql}</code></pre>
                """

            logger.friendlyHTML(detailed_explanation)

    def display(self) -> None:
        """Executes the data exploration queries and displays a sample of results"""
        return self.execute()

    def execute(self) -> None:
        """[DEPRECATED] Executes the data exploration queries and displays a sample of results"""
        df = self.to_union_dataframe()
        try:
            df.display()
        except Exception as e:
            df.show(truncate=False)

    def apply(self) -> DataFrame:
        """Executes the data exploration queries and returns a DataFrame with the results"""
        return self.to_union_dataframe()

    def to_union_dataframe(self) -> DataFrame:
        """[DEPRECATED] Executes the data exploration queries and returns a DataFrame with the results"""

        sql_commands = self._get_sql_commands(self._data_explorer)
        dfs = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=self._data_explorer._max_concurrency) as executor:
            # Submit tasks to the thread pool
            futures = [executor.submit(self._run_sql, sql, table) for sql, table in sql_commands]

            # Process completed tasks
            for future in concurrent.futures.as_completed(futures):
                result = future.result()
                if result is not None:
                    dfs.append(result)

        logger.debug("Finished lakehouse scanning task")

        if dfs:
            return reduce(lambda x, y: x.union(y), dfs)
        else:
            raise Exception("No SQL statements were successfully executed.")
